#include "squidwall.h"
#include "hextab.h"
#include "md5.h"

/**
 * Copyright (C) 2006 Tino Reichardt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License Version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/**
 * The cache is implemented as a static array with N members.
 * - the cache isn't ordered in any manner
 * - we do a linear search for some item
 *   - on this search, we remember the oldest entry and delete it later (if the
 *     cache size is reached)
 *   - we may also use the hitcounter for this
 */

static array array_memcache;

/**
 * memcache_get_hash - generate hash value of the squid variables
 *
 * @return the null-terminated md5 hash of the variables
 */
static char *memcache_get_hash(void)
{
	unsigned char digest[16];
	static char hash[33];
	char *e;
	MD5_CTX md5;
	int i;

	MD5_Init(&md5);
	MD5_Update(&md5, squid_url,    str_len(squid_url));
	MD5_Update(&md5, squid_ip,     str_len(squid_ip));
	MD5_Update(&md5, squid_host,   str_len(squid_host));
	MD5_Update(&md5, squid_user,   str_len(squid_user));
	MD5_Update(&md5, squid_method, str_len(squid_method));
	MD5_Final(digest, &md5);

	e=hash;
	for (i=0; i<16; i++) {
		*e = hextab[digest[i]/16]; ++e;
		*e = hextab[digest[i]%16]; ++e;
	}
	*e=0;

	return hash;
}


/**
 * memcache_init - allocates memory for the cache
 */
void memcache_init(void)
{
	struct memcache *mc;
	int i;

	debug_s(DEBUG_MEMCACHE, "memcache_init()");
	mc=array_allocate(&array_memcache, sizeof(struct memcache),
			memcache*sizeof(struct memcache));
	if (!mc) die_nomem();
	for (i=0; i < array_length(&array_memcache, sizeof(struct memcache)); i++) {
		mc = array_get(&array_memcache, sizeof(struct memcache), i);
		byte_zero(mc->hash, sizeof(mc->hash));
		stralloc_init(&mc->url);
		mc->hits=0; mc->ts=0;
	}

	return;
}


/**
 * memcache_insert - insert some new entry in our small memory cache
 *
 * @param result - gets filled with the url to be redirected to
 * @return: MEMCACHE_MISS there was no such entry in the cache
 *          MEMCACHE_HIT  the current url was found, result is set
 */
void memcache_insert(void)
{
	struct memcache *mc=0;
	time_t oldest=0;
	int i, id_old=0;
	char *hash;

	hash = memcache_get_hash();
	mc = array_get(&array_memcache, sizeof(struct memcache), 0);
	for (i=0; i < array_length(&array_memcache, sizeof(struct memcache)); i++) {
		mc = array_get(&array_memcache, sizeof(struct memcache), i);

		if (mc->ts == 0) break; /* last entry */

		/* check, if this is the oldest entry */
		if (oldest) {
			if (mc->ts > oldest) oldest = mc->ts;
		} else {
			oldest = mc->ts;
			id_old=i;
		}

		if (str_equal(mc->hash, hash)) {
			debug_s3(DEBUG_MEMCACHE, "memcache_insert() REFRESH OLD [", hash, "]");
			if (!stralloc_copy(&mc->url, &redirsite)) die_nomem();
			mc->ts=time(0); mc->hits++;
			return;
		}
	}

	/* add new entry @end of list */
	if (mc->ts == 0) {
		debug_s3(DEBUG_MEMCACHE, "memcache_insert() ADD @ END [", hash, "]");
		if (!stralloc_copy(&mc->url, &redirsite)) die_nomem();
		mc->hash[fmt_str(mc->hash, hash)]=0;
		mc->ts=time(0);
	} else {
		mc = array_get(&array_memcache, sizeof(struct memcache), id_old);
		debug_s3(DEBUG_MEMCACHE, "memcache_insert() REPLACE OLD [", hash, "]");
		if (!stralloc_copy(&mc->url, &redirsite)) die_nomem();
		mc->hash[fmt_str(mc->hash, hash)]=0;
		mc->ts=time(0);
	}

	return;
}


/**
 * memcache_quit - free the memory cache
 */
void memcache_quit(void)
{
	struct memcache *mc;
	int i;

	debug_s(DEBUG_MEMCACHE, "memcache_quit()");
	for (i=0; i < array_length(&array_memcache, sizeof(struct memcache)); i++) {
		mc = array_get(&array_memcache, sizeof(struct memcache), i);
		stralloc_free(&mc->url);
	}
	array_reset(&array_memcache);

	return;
}


/**
 * memcache_lookup - look up some entry in our small memory cache
 *
 * @param result gets filled with the url to be redirected to
 * @return: MEMCACHE_MISS there was no such entry in the cache
 *          MEMCACHE_HIT  the current url was found, result is set
 */
int memcache_lookup(void)
{
	struct memcache *mc;
	char *hash;
	int i;

	hash = memcache_get_hash();
	for (i=0; i < array_length(&array_memcache, sizeof(struct memcache)); i++) {
		mc = array_get(&array_memcache, sizeof(struct memcache), i);
		if (mc->ts == 0) break; /* last entry */
		if (str_equal(mc->hash, hash)) {
			log_headline("memcache");
			log_s("MEMCACHE_HIT ["); log_s(squid_url); log_s("]\n");
			debug_s3(DEBUG_MEMCACHE, "memcache_lookup() MEMCACHE_HIT [", squid_url, "]");
			if (!stralloc_cat(&redirsite, &mc->url)) die_nomem();
			mc->ts=time(0); mc->hits++;
			return MEMCACHE_HIT;
		}
	}

	debug_s(DEBUG_MEMCACHE, "memcache_lookup() MEMCACHE_MISS");
	return MEMCACHE_MISS;
}
